@@ -3,6 +3,8 @@
 package Unicode::LineBreak;
 require 5.008;
 
+=encoding utf8
+
 =head1 NAME
 
 Unicode::LineBreak - UAX #14 Unicode Line Breaking Algorithm
@@ -41,31 +43,26 @@ use MIME::Charset qw(:info);
 ### Globals
 
 ### The package version, both in 1.23 style *and* usable by MakeMaker:
-$VERSION = '0.001_01';
+$VERSION = '0.001_02';
 
 ### Public Configuration Attributes
-$Config = {
+our $Config = {
     %{$MIME::Charset::Config}, # Detect7bit, Replacement, Mapping
+    Break => "\n",
     Charset => 'UTF-8',
     Context => '',
     HangulAsAL => 'NO',
     Language => 'XX',
+    LegacyCM => 'YES',
     MaxColumns => 76,
     NSKanaAsID => 'NO',
     OutputCharset => 'UTF-8',
 };
 eval { require Unicode::LineBreak::Defaults; };
 
-### Constants
-use constant MANDATORY => 2;
-use constant ALLOWED => 1;
-use constant NO_BREAK => 0;
-use constant EOT => -1;
-
 ### Privates
-require Unicode::LineBreak::Data;
 require Unicode::LineBreak::Rules;
-sub lb_null { "" }
+require Unicode::LineBreak::Data;
 
 my @ALL_CLASSES = qw(BK CR LF CM NL SG WJ ZW GL SP B2 BA BB HY CB CL EX IN NS OP QU IS NU PO PR SY AI AL H2 H3 ID JL JV JT SA XX);  
 
@@ -92,7 +89,7 @@ my $EASTASIAN_LANGUAGES = qr{
 
 =over 4
 
-=item new STRING, [OPTIONS...]
+=item new STRING, [OPTIONS, ...]
 
 Constructor.  Following OPTIONS may be specified.
 
@@ -103,27 +100,39 @@ Constructor.  Following OPTIONS may be specified.
 Character set that is used to encode string STRING.
 Default is C<"UTF-8">.
 
-=item Context => C<"EASTASIAN"> | C<"NONEASTASIAN">
+=item Context => CONTEXT
 
-=item Language => LANGUAGE
-
-Along with Charset option, these options may be used to define
-language/region context.
+Along with Charset option, this may be used to define language/region
+context.
 Currently available contexts are C<"EASTASIAN"> and C<"NONEASTASIAN">.
+Default context is C<"NONEASTASIAN">.
 
 =item HangulAsAL => C<"YES"> | C<"NO">
 
 Treat hangul syllables and conjoining jamos as alphabetic characters (AL).
 Default is C<"NO">.
 
+=item Language => LANGUAGE
+
+Along with Charset option, this may be used to define language/region
+context.
+See Context option.
+
+=item LegacyCM => C<"YES"> | C<"NO">
+
+Treat combining characters lead by SPACE as an isolated combining character.
+As of Unicode 5.0, such use of SPACE is not recommended.
+Default is C<"YES">.
+
 =item MaxColumns => NUMBER
 
-Maximum number of columns line may include, in other words, length of line.
+Maximum number of columns line may include not counting trailing spaces and
+newline sequence.  In other words, maximum length of line.
 Default is C<76>.
 
 =item NSKanaAsID => C<"YES"> | C<"NO">
 
-Treat hiragana/katakana non-starters and prolonged signs (NS) as
+Treat hiragana/katakana small letters and prolonged signs (NS) as
 ideographic characters (ID).
 This feature is optional in [JIS X 4051].
 Default is C<"NO">.
@@ -146,7 +155,6 @@ sub new {
 
     my $self = { };
     &config($self, @_);
-    &setRules($self);
     if (Encode::is_utf8($str)) {
 	$self->{_str} = $str;
     } else {
@@ -171,6 +179,7 @@ sub config {
     }
     $self->{_language} = uc($params{Language} || $Config->{Language});
     $self->{_language} =~ s/_/-/g;
+    $self->{_legacy_cm} = uc($params{LegacyCM} || $Config->{LegacyCM});
     $self->{_max_columns} = $params{MaxColumns} || $Config->{MaxColumns};
 
     # Context. Either East Asian or Non-East Asian.
@@ -191,132 +200,184 @@ sub config {
     $self->{_context} = $context;
 
     # Some flags
-    my $nskanaasid = uc($params{NSKanaAsID} || $Config->{NSKanaAsID});
-    my $hangulasal = uc($params{HangulAsAL} || $Config->{HangulAsAL});
+    $self->{_ns_kana_as_id} = uc($params{NSKanaAsID} || $Config->{NSKanaAsID});
+    $self->{_hangul_as_al} = uc($params{HangulAsAL} || $Config->{HangulAsAL});
+    $self->{_break_sequence} = "\n";
 
     ## Customize Line Breaking Classes.
 
     my %lb;
-    foreach my $c (@ALL_CLASSES) {
+    foreach my $c (qw{BK CR LF NL SP ZW CM H2 H3 JL JV JT}) {
 	$lb{$c} = [$c];
     }
-
-    # Resolve SA, SG and XX: See UAX #14 6.1 LB1.
-    push @{$lb{CM}}, 'SAcm';
-    push @{$lb{AL}}, qw(SA SG XX);
-
-    # Resolve AI.
-    if ($context eq 'EASTASIAN') {
-	push @{$lb{ID}}, 'AI';
-    } else {
-	push @{$lb{AL}}, 'AI';
-    }
-
-    # Options for katakana/hiragana NS: See [JIS X 4051] 6.1.1 note 8.
-    if ($nskanaasid eq 'YES') {
-	push @{$lb{ID}}, 'NSid';
-    } else {
-	push @{$lb{NS}}, 'NSid';
-    }
-
-    # Options for hangul syllables and conjoining jamos: See [UAX #14] 5.1.
-    if ($hangulasal eq 'YES') {
-	foreach my $c (qw(H2 H3 JL JT JV)) {
-	    push @{$lb{AL}}, $c;
-	    $lb{$c} = [qw(null)];
-	}
-    }
-
+    push @{$lb{CM}}, 'SAcm'; # Resolve SA: See UAX #14 6.1 LB1.
+    $lb{SA} = [qw{SAcm SAal}];
+    $lb{hangul} = [qw{H3 H2 JL JV JT}];
     foreach my $c (keys %lb) {
 	$self->{"lb_$c"} = '(?:' .
 	    (join '|', map qr{\p{lb_$_}}, @{$lb{$c}}) .
 	    ')';
     }
-
-    ## Customize East Asian Widths.
-
-    $self->{"ea_z"} = qr{\p{ea_z}}os;
-    if ($context eq 'EASTASIAN') {
-	$self->{"ea_w"} = qr{(?:\p{ea_F}|\p{ea_W}|\p{ea_A})}os;
-	$self->{"ea_n"} = qr{(?:\p{ea_H}|\p{ea_Na}|\p{ea_N})}os;
-    } else {
-	$self->{"ea_w"} = qr{(?:\p{ea_F}|\p{ea_W})}os;
-	$self->{"ea_n"} = qr{(?:\p{ea_H}|\p{ea_Na}|\p{ea_A}|\p{ea_N})}os;
-    }
 }
 
 =over 4
 
 =item break
 
-Break string and returns it.
+Instance method.  Break string and returns it.
 
 =back
 
 =cut
 
 sub break {
-    my $self = shift;
-    my ($action, $sym);
-    my $result = '';
-    my $bbuf = '';
-    my $blen = 0;
-    my $pbuf = '';
-    my $plen = 0;
-    my $slen;
-    my $sp;
-    my $str = $self->{_str};
-
+    my $s = shift;
+    my $str = $s->{_str};
     return '' unless defined $str and length $str;
 
+    my $result = '';
+    my ($l_frag, $l_len) = ('', 0);
+    my ($b_frag, $b_spc, $b_cls) = ('', '', undef);
     pos($str) = 0;
     while (1) {
-	foreach my $r (@{$self->{_rules}}) {
-	    if ($str =~ m/$r->[0]/cg) {
-		($action, $sym) = ($r->[1], $&);
-		last;
+	my ($a_frag, $a_spc, $a_cls);
+
+	# LB3: × eot
+	if ($str =~ /\G\z/cgs) {
+	    $b_cls = 'eot';
+	    ($a_frag, $a_spc, $a_cls) = ('', '', undef);
+	# LB5, LB6: × (BK | CR LF | CR | LF | NL) !
+	} elsif ($str =~
+		 /\G(?:$s->{lb_BK}|
+		     $s->{lb_CR}$s->{lb_LF}|
+		     $s->{lb_CR}|
+		     $s->{lb_LF}|
+		     $s->{lb_NL})/cgsx) {
+	    $b_frag .= $b_spc.$&; #FIXME: Process mandatory break.
+	    $b_spc = '';
+	    $b_cls = 'BK';
+	    ($a_frag, $a_spc, $a_cls) = ('', '', undef);
+	# LB7, LB8: × (ZW | SP)* ZW 
+	} elsif ($str =~ /\G(?:$s->{lb_ZW}|$s->{lb_SP})*$s->{lb_ZW}/cgs) {
+	    $b_frag .= $b_spc.$&;
+	    $b_spc = '';
+	    $b_cls = 'ZW';
+	    next;
+	# LB7: × SP+
+	} elsif ($str =~ /\G$s->{lb_SP}+/cgs) {
+	    $b_spc .= $&;
+	    $b_cls ||= 'WJ'; # in case of --- (sot | BK etc. | ZW) × SP+
+	    next;
+	# LB7, LB9, LB26, LB27: Treat
+	#   (JL* H3 JT* | JL* H2 JV* JT* | JL* JV+ JT* | JL+ | JT+) CM* SP*
+	# as if it were ID CM* SP* (or optionally AL CM* SP*)
+	} elsif ($str =~
+		 /\G((?:
+		      $s->{lb_JL}* $s->{lb_H3} $s->{lb_JT}* |
+		      $s->{lb_JL}* $s->{lb_H2} $s->{lb_JV}* $s->{lb_JT}* |
+		      $s->{lb_JL}* $s->{lb_JV}+ $s->{lb_JT}* |
+		      $s->{lb_JL}+ | $s->{lb_JT}+)
+		     $s->{lb_CM}*) ($s->{lb_SP}*)/cgsx) {
+	    ($a_frag, $a_spc) = ($1, $2);
+	    if ($s->{_hangul_as_al} eq 'YES') {
+		$a_cls = 'AL';
+	    } else {
+		$a_cls = 'ID';
+	    }
+	# LB7, LB9: Treat X CM* SP* as if it were X SP*
+	# where X is anything except BK, CR, LF, NL, SP or ZW
+	} elsif ($str =~ /\G(.$s->{lb_CM}*)($s->{lb_SP}*)/cgs) {
+	    ($a_frag, $a_spc) = ($1, $2);
+
+	    # LB1: Assign a line breaking class to each characters.
+	    $a_cls = &_bsearch($Unicode::LineBreak::lb_MAP, $a_frag) || 'XX';
+	    $a_cls = {
+		'SAcm' => 'CM',
+		'SAal' => 'AL',
+		'SG' => 'AL',
+		'XX' => 'AL',
+		'AI' => ($s->{_context} eq 'EASTASIAN'? 'ID': 'AL'),
+		'NSid' => ($s->{_ns_kana_as_id} eq 'YES'? 'ID': 'NS'),
+	    }->{$a_cls} || $a_cls;
+
+	    if ($a_cls eq 'CM') {
+		# LB7, Legacy-CM: Treat SP CM+ SP* as if it were ID SP*
+		# See [UAX #14] 9.1.
+		if ($s->{_legacy_cm} and $b_spc =~ s/.$//os) {
+		    $a_frag = $&.$a_frag;
+		    $a_cls = 'ID';
+		    $b_cls = undef unless length $b_frag.$b_spc; # clear
+		# LB7, LB10: Treat CM+ SP* as if it were AL SP*
+		} else {
+		    $a_cls = 'AL';
+		}
+	    }
+	} else {
+	    croak pos($str).": This should not happen: ask the developer.";
+	}
+
+	# LB2: sot ×
+	unless ($b_cls) {
+	    ($b_frag, $b_spc, $b_cls) =	($a_frag, $a_spc, $a_cls);
+	    next;
+	}
+
+	my $action;
+	if ($b_cls eq 'eot') {
+	    $action = 'EOT';
+	# LB4, LB5: (BK | CR LF | CR | LF | NL) !
+	} elsif ($b_cls eq 'BK') {
+	    $action = 'MANDATORY';
+	# LB11 - LB29 and LB31: Tailorable rules (except LB11).
+	} else {
+	    my $b_idx = $Unicode::LineBreak::lb_IDX{$b_cls};
+	    my $a_idx = $Unicode::LineBreak::lb_IDX{$a_cls};
+	    $action = $Unicode::LineBreak::RULES_MAP->[$b_idx]->[$a_idx];
+	    # LB31: ALL ÷ ALL
+	    $action ||=	'DIRECT';
+	    # Resolve indirect break.
+	    $action = 'PROHIBITED' if $action eq 'INDIRECT' and !length $b_spc;
+
+	    if ($action eq 'PROHIBITED') {
+		$b_frag .= $b_spc.$a_frag;
+		$b_spc = $a_spc;
+		$b_cls = $a_cls;
+		next;
 	    }
 	}
-	if ($sym =~ s/\u0020+$//) {
-	    $sp = $&;
+
+	if ($s->{_max_columns} < $l_len + $s->_strwidth($b_frag)) {
+            #FIXME: Process arbitrary break
+	    $result .= $l_frag."\n" if length $l_frag;
+	    $l_frag = $b_frag.$b_spc;
+	    $l_len = $s->_strwidth($b_frag.$b_spc);
 	} else {
-	    $sp = '';
+	    $l_frag .= $b_frag.$b_spc;
+	    $l_len += $s->_strwidth($b_frag.$b_spc);
 	}
-	$slen = $self->_strwidth($sym);
+	($b_frag, $b_spc, $b_cls) = ($a_frag, $a_spc, $a_cls);
 
-	if ($bbuf ne '' and $self->{_max_columns} < $blen + $plen + $slen) {
-	    $result .= $bbuf."\n";
-	    $bbuf = '';
-	    $blen = 0;
-	}
-	$plen += $slen;
-	if ($slen and $pbuf =~ /\u0020+$/) {
-	    $plen += length $&;
-	}
-	$pbuf .= $sym.$sp;
-
-	if ($action == EOT) {
-	    $result .= $bbuf.$pbuf;
+	if ($action eq 'MANDATORY') {
+	    $result .= $l_frag;
+	    $l_frag = '';
+	    $l_len = 0;
+	} elsif ($action eq 'EOT') {
+	    $result .= $l_frag;
 	    last;
-	} elsif ($action == MANDATORY) {
-	    $result .= $bbuf.$pbuf;
-	    $bbuf = $pbuf = '';
-	    $blen = $plen = 0;
-	} elsif ($action == ALLOWED) {
-	    $bbuf .= $pbuf;
-	    $blen += $plen;
-	    $pbuf = '';
-	    $plen = 0;
 	}
     }
 
-    if ($self->{_output_charset} eq '_UNICODE_') {
+    if ($s->{_output_charset} eq '_UNICODE_') {
 	return $result;
     } else {
-	return $self->{_charset}->encode($result);
+	return $s->{_charset}->encode($result);
     }
 }
 
+# Helper functions.
+
+# self->_strwidth STR
+# Coliculate the number of columns that string STR will occupy.
 # TODO: Adjust widths of hangul conjoining jamos.
 sub _strwidth {
     my $self = shift;
@@ -325,17 +386,50 @@ sub _strwidth {
 
     return 0 unless defined $str and length $str;
     my $i;
-    for ($i = 0; $i < length $str; $i++) {
-	my $char = substr($str, $i, 1);
-	if ($char =~ /$self->{ea_w}/) {
+    for ($i = 0; $i < length($str); $i++) {
+	my $width = &_bsearch($Unicode::LineBreak::ea_MAP,
+			      substr($str, $i, 1));
+	if ($width eq 'F' or $width eq 'W') {
 	    $result += 2;
-	} elsif ($char !~ /$self->{ea_z}/) {
+        } elsif ($self->{_context} eq 'EASTASIAN' and $width eq 'A') {
+            $result += 2;
+	} elsif ($width ne 'z') {
 	    $result += 1;
 	}
     }
     return $result;
 }
 
+# _bearch MAP, VAL
+# Examine binary search on property map table with following structure:
+# [
+#     [start, stop, property_value],
+#     ...
+# ]
+# where start and stop stands for a continuous range of UCS ordinal those
+# are assigned property_value.
+sub _bsearch {
+    my $map = shift;
+    my $val = ord shift;
+
+    my $top = 0;
+    my $bot = $#{$map};
+    my $cur;
+
+    while ($top <= $bot) {
+	$cur = $top + int(($bot - $top) / 2);
+	my $v = $map->[$cur];
+	if ($val < $v->[0]) {
+	    $bot = $cur - 1;
+	} elsif ($v->[1] < $val) {
+	    $top = $cur + 1;
+	} else {
+	    return $v->[2];
+	}
+    }
+    return undef;
+}
+
 =head2 Configuration Files
 
 Built-in defaults of option parameters for L<"new"> method
@@ -381,9 +475,13 @@ Characters assigned to SG or XX are resolved to AL.
 
 =back
 
+=head1 CAVEAT
+
+I<To be written>.
+
 =head1 BUGS
 
-B<Slightly slow>.  This is pre-alpha release for proof-of-concept.
+Please report bugs or buggy behaviors to developer.  See L</AUTHOR>.
 
 =head1 VERSION
 
@@ -430,7 +528,7 @@ L<http://unicode.org/reports/tr14/>.
 
 =back
 
-=head1 AUTHORS
+=head1 AUTHOR
 
 Copyright (C) 2009 Hatuka*nezumi - IKEDA Soji <hatuka(at)nezumi.nu>.
 
